@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveFunctor #-}
 {-# LANGUAGE DeriveDataTypeable         #-}
 {-# LANGUAGE DeriveGeneric #-}
 {-# LANGUAGE TupleSections #-}
@@ -9,8 +10,7 @@
 {-# LANGUAGE UndecidableInstances #-}
 module Yesod.Core.Types where
 
-import qualified Blaze.ByteString.Builder           as BBuilder
-import qualified Blaze.ByteString.Builder.Char.Utf8
+import qualified Data.ByteString.Builder            as BB
 #if __GLASGOW_HASKELL__ < 710
 import           Control.Applicative                (Applicative (..))
 import           Control.Applicative                ((<$>))
@@ -18,21 +18,20 @@ import           Data.Monoid                        (Monoid (..))
 #endif
 import           Control.Arrow                      (first)
 import           Control.Exception                  (Exception)
-import           Control.Monad                      (liftM, ap)
-import           Control.Monad.Base                 (MonadBase (liftBase))
-import           Control.Monad.Catch                (MonadMask (..), MonadCatch (..))
+import           Control.Monad                      (ap)
 import           Control.Monad.IO.Class             (MonadIO (liftIO))
 import           Control.Monad.Logger               (LogLevel, LogSource,
                                                      MonadLogger (..))
-import           Control.Monad.Trans.Control        (MonadBaseControl (..))
-import           Control.Monad.Trans.Resource       (MonadResource (..), InternalState, runInternalState, MonadThrow (..), monadThrow, ResourceT)
+import           Control.Monad.Trans.Resource       (MonadResource (..), InternalState, runInternalState, MonadThrow (..), ResourceT)
 import           Data.ByteString                    (ByteString)
 import qualified Data.ByteString.Lazy               as L
-import           Data.Conduit                       (Flush, Source)
-import           Data.IORef                         (IORef)
+import           Data.CaseInsensitive               (CI)
+import           Data.Conduit                       (Flush, ConduitT)
+import           Data.IORef                         (IORef, modifyIORef')
 import           Data.Map                           (Map, unionWith)
 import qualified Data.Map                           as Map
 import           Data.Monoid                        (Endo (..), Last (..))
+import           Data.Semigroup                     (Semigroup(..))
 import           Data.Serialize                     (Serialize (..),
                                                      putByteString)
 import           Data.String                        (IsString (fromString))
@@ -49,27 +48,19 @@ import           Network.Wai                        (FilePart,
 import qualified Network.Wai                        as W
 import qualified Network.Wai.Parse                  as NWP
 import           System.Log.FastLogger              (LogStr, LoggerSet, toLogStr, pushLogStr)
-import qualified System.Random.MWC                  as MWC
 import           Network.Wai.Logger                 (DateCacheGetter)
 import           Text.Blaze.Html                    (Html, toHtml)
 import           Text.Hamlet                        (HtmlUrl)
 import           Text.Julius                        (JavascriptUrl)
 import           Web.Cookie                         (SetCookie)
 import           Yesod.Core.Internal.Util           (getTime, putTime)
-import           Control.Monad.Trans.Class          (MonadTrans (..))
 import           Yesod.Routes.Class                 (RenderRoute (..), ParseRoute (..))
 import           Control.Monad.Reader               (MonadReader (..))
-#if !MIN_VERSION_base(4, 6, 0)
-import Prelude hiding (catch)
-#endif
 import Control.DeepSeq (NFData (rnf))
 import Control.DeepSeq.Generics (genericRnf)
-import Data.Conduit.Lazy (MonadActive, monadActive)
 import Yesod.Core.TypeCache (TypeMap, KeyedTypeMap)
-#if MIN_VERSION_monad_logger(0, 3, 10)
 import Control.Monad.Logger (MonadLoggerIO (..))
-#endif
-import Data.Semigroup (Semigroup)
+import UnliftIO (MonadUnliftIO (..), UnliftIO (..))
 
 -- Sessions
 type SessionMap = Map Text ByteString
@@ -82,7 +73,7 @@ newtype SessionBackend = SessionBackend
                     -> IO (SessionMap, SaveSession) -- ^ Return the session data and a function to save the session
     }
 
-data SessionCookie = SessionCookie (Either UTCTime ByteString) ByteString SessionMap
+data SessionCookie = SessionCookie !(Either UTCTime ByteString) !ByteString !SessionMap
     deriving (Show, Read)
 instance Serialize SessionCookie where
     put (SessionCookie a b c) = do
@@ -140,13 +131,13 @@ type RequestBodyContents =
 data FileInfo = FileInfo
     { fileName        :: !Text
     , fileContentType :: !Text
-    , fileSourceRaw   :: !(Source (ResourceT IO) ByteString)
+    , fileSourceRaw   :: !(ConduitT () ByteString (ResourceT IO) ())
     , fileMove        :: !(FilePath -> IO ())
     }
 
 data FileUpload = FileUploadMemory !(NWP.BackEnd L.ByteString)
                 | FileUploadDisk !(InternalState -> NWP.BackEnd FilePath)
-                | FileUploadSource !(NWP.BackEnd (Source (ResourceT IO) ByteString))
+                | FileUploadSource !(NWP.BackEnd (ConduitT () ByteString (ResourceT IO) ()))
 
 -- | How to determine the root of the application for constructing URLs.
 --
@@ -160,13 +151,13 @@ data Approot master = ApprootRelative -- ^ No application root.
 
 type ResolvedApproot = Text
 
-data AuthResult = Authorized | AuthenticationRequired | Unauthorized Text
+data AuthResult = Authorized | AuthenticationRequired | Unauthorized !Text
     deriving (Eq, Show, Read)
 
 data ScriptLoadPosition master
     = BottomOfBody
     | BottomOfHeadBlocking
-    | BottomOfHeadAsync (BottomOfHeadAsync master)
+    | BottomOfHeadAsync !(BottomOfHeadAsync master)
 
 type BottomOfHeadAsync master
        = [Text] -- ^ urls to load asynchronously
@@ -179,14 +170,16 @@ type Texts = [Text]
 newtype WaiSubsite = WaiSubsite { runWaiSubsite :: W.Application }
 
 -- | Like 'WaiSubsite', but applies parent site's middleware and isAuthorized.
--- 
+--
 -- @since 1.4.34
 newtype WaiSubsiteWithAuth = WaiSubsiteWithAuth { runWaiSubsiteWithAuth :: W.Application }
 
-data RunHandlerEnv site = RunHandlerEnv
+data RunHandlerEnv child site = RunHandlerEnv
     { rheRender   :: !(Route site -> [(Text, Text)] -> Text)
-    , rheRoute    :: !(Maybe (Route site))
+    , rheRoute    :: !(Maybe (Route child))
+    , rheRouteToMaster :: !(Route child -> Route site)
     , rheSite     :: !site
+    , rheChild    :: !child
     , rheUpload   :: !(RequestBodyLength -> FileUpload)
     , rheLog      :: !(Loc -> LogSource -> LogLevel -> LogStr -> IO ())
     , rheOnError  :: !(ErrorResponse -> YesodApp)
@@ -196,11 +189,10 @@ data RunHandlerEnv site = RunHandlerEnv
     , rheMaxExpires :: !Text
     }
 
-data HandlerData site parentRoute = HandlerData
+data HandlerData child site = HandlerData
     { handlerRequest  :: !YesodRequest
-    , handlerEnv      :: !(RunHandlerEnv site)
+    , handlerEnv      :: !(RunHandlerEnv child site)
     , handlerState    :: !(IORef GHState)
-    , handlerToParent :: !(Route site -> parentRoute)
     , handlerResource :: !InternalState
     }
 
@@ -208,40 +200,38 @@ data YesodRunnerEnv site = YesodRunnerEnv
     { yreLogger         :: !Logger
     , yreSite           :: !site
     , yreSessionBackend :: !(Maybe SessionBackend)
-    , yreGen            :: !MWC.GenIO
-    , yreGetMaxExpires  :: IO Text
+    , yreGen            :: !(IO Int)
+    -- ^ Generate a random number
+    , yreGetMaxExpires  :: !(IO Text)
     }
 
-data YesodSubRunnerEnv sub parent parentMonad = YesodSubRunnerEnv
-    { ysreParentRunner  :: !(ParentRunner parent parentMonad)
+data YesodSubRunnerEnv sub parent = YesodSubRunnerEnv
+    { ysreParentRunner  :: !(ParentRunner parent)
     , ysreGetSub        :: !(parent -> sub)
     , ysreToParentRoute :: !(Route sub -> Route parent)
     , ysreParentEnv     :: !(YesodRunnerEnv parent) -- FIXME maybe get rid of this and remove YesodRunnerEnv in ParentRunner?
     }
 
-type ParentRunner parent m
-    = m TypedContent
+type ParentRunner parent
+    = HandlerFor parent TypedContent
    -> YesodRunnerEnv parent
    -> Maybe (Route parent)
    -> W.Application
 
 -- | A generic handler monad, which can have a different subsite and master
 -- site. We define a newtype for better error message.
-newtype HandlerT site m a = HandlerT
-    { unHandlerT :: HandlerData site (MonadRoute m) -> m a
+newtype HandlerFor site a = HandlerFor
+    { unHandlerFor :: HandlerData site site -> IO a
     }
-
-type family MonadRoute (m :: * -> *)
-type instance MonadRoute IO = ()
-type instance MonadRoute (HandlerT site m) = (Route site)
+    deriving Functor
 
 data GHState = GHState
-    { ghsSession :: SessionMap
-    , ghsRBC     :: Maybe RequestBodyContents
-    , ghsIdent   :: Int
-    , ghsCache   :: TypeMap
-    , ghsCacheBy :: KeyedTypeMap
-    , ghsHeaders :: Endo [Header]
+    { ghsSession :: !SessionMap
+    , ghsRBC     :: !(Maybe RequestBodyContents)
+    , ghsIdent   :: !Int
+    , ghsCache   :: !TypeMap
+    , ghsCacheBy :: !KeyedTypeMap
+    , ghsHeaders :: !(Endo [Header])
     }
 
 -- | An extension of the basic WAI 'W.Application' datatype to provide extra
@@ -252,24 +242,35 @@ type YesodApp = YesodRequest -> ResourceT IO YesodResponse
 -- | A generic widget, allowing specification of both the subsite and master
 -- site datatypes. While this is simply a @WriterT@, we define a newtype for
 -- better error messages.
-newtype WidgetT site m a = WidgetT
-    { unWidgetT :: HandlerData site (MonadRoute m) -> m (a, GWData (Route site))
+newtype WidgetFor site a = WidgetFor
+    { unWidgetFor :: WidgetData site -> IO a
     }
+    deriving Functor
 
-instance (a ~ (), Monad m) => Monoid (WidgetT site m a) where
+data WidgetData site = WidgetData
+  { wdRef :: {-# UNPACK #-} !(IORef (GWData (Route site)))
+  , wdHandler :: {-# UNPACK #-} !(HandlerData site site)
+  }
+
+instance a ~ () => Monoid (WidgetFor site a) where
     mempty = return ()
-    mappend x y = x >> y
-instance (a ~ (), Monad m) => Semigroup (WidgetT site m a)
+#if !(MIN_VERSION_base(4,11,0))
+    mappend = (<>)
+#endif
+instance a ~ () => Semigroup (WidgetFor site a) where
+    x <> y = x >> y
 
 -- | A 'String' can be trivially promoted to a widget.
 --
 -- For example, in a yesod-scaffold site you could use:
 --
 -- @getHomeR = do defaultLayout "Widget text"@
-instance (Monad m, a ~ ()) => IsString (WidgetT site m a) where
+instance a ~ () => IsString (WidgetFor site a) where
     fromString = toWidget . toHtml . T.pack
-      where toWidget x = WidgetT $ const $ return ((), GWData (Body (const x))
-                         mempty mempty mempty mempty mempty mempty)
+      where toWidget x = tellWidget mempty { gwdBody = Body (const x) }
+
+tellWidget :: GWData (Route site) -> WidgetFor site ()
+tellWidget d = WidgetFor $ \wd -> modifyIORef' (wdRef wd) (<> d)
 
 type RY master = Route master -> [(Text, Text)] -> Text
 
@@ -287,13 +288,13 @@ newtype CssBuilder = CssBuilder { unCssBuilder :: TBuilder.Builder }
 --
 -- > PageContent url -> HtmlUrl url
 data PageContent url = PageContent
-    { pageTitle :: Html
-    , pageHead  :: HtmlUrl url
-    , pageBody  :: HtmlUrl url
+    { pageTitle :: !Html
+    , pageHead  :: !(HtmlUrl url)
+    , pageBody  :: !(HtmlUrl url)
     }
 
-data Content = ContentBuilder !BBuilder.Builder !(Maybe Int) -- ^ The content and optional content length.
-             | ContentSource !(Source (ResourceT IO) (Flush BBuilder.Builder))
+data Content = ContentBuilder !BB.Builder !(Maybe Int) -- ^ The content and optional content length.
+             | ContentSource !(ConduitT () (Flush BB.Builder) (ResourceT IO) ())
              | ContentFile !FilePath !(Maybe FilePart)
              | ContentDontEvaluate !Content
 
@@ -316,11 +317,11 @@ newtype DontFullyEvaluate a = DontFullyEvaluate { unDontFullyEvaluate :: a }
 -- | Responses to indicate some form of an error occurred.
 data ErrorResponse =
       NotFound
-    | InternalError Text
-    | InvalidArgs [Text]
+    | InternalError !Text
+    | InvalidArgs ![Text]
     | NotAuthenticated
-    | PermissionDenied Text
-    | BadMethod H.Method
+    | PermissionDenied !Text
+    | BadMethod !H.Method
     deriving (Show, Eq, Typeable, Generic)
 instance NFData ErrorResponse where
     rnf = genericRnf
@@ -328,9 +329,11 @@ instance NFData ErrorResponse where
 ----- header stuff
 -- | Headers to be added to a 'Result'.
 data Header =
-      AddCookie SetCookie
-    | DeleteCookie ByteString ByteString
-    | Header ByteString ByteString
+      AddCookie !SetCookie
+    | DeleteCookie !ByteString !ByteString
+    -- ^ name and path
+    | Header !(CI ByteString) !ByteString
+    -- ^ key and value
     deriving (Eq, Show)
 
 -- FIXME In the next major version bump, let's just add strictness annotations
@@ -341,25 +344,27 @@ instance NFData Header where
     rnf (DeleteCookie x y) = x `seq` y `seq` ()
     rnf (Header x y) = x `seq` y `seq` ()
 
-data Location url = Local url | Remote Text
+data Location url = Local !url | Remote !Text
     deriving (Show, Eq)
 
 -- | A diff list that does not directly enforce uniqueness.
 -- When creating a widget Yesod will use nub to make it unique.
 newtype UniqueList x = UniqueList ([x] -> [x])
 
-data Script url = Script { scriptLocation :: Location url, scriptAttributes :: [(Text, Text)] }
+data Script url = Script { scriptLocation :: !(Location url), scriptAttributes :: ![(Text, Text)] }
     deriving (Show, Eq)
-data Stylesheet url = Stylesheet { styleLocation :: Location url, styleAttributes :: [(Text, Text)] }
+data Stylesheet url = Stylesheet { styleLocation :: !(Location url), styleAttributes :: ![(Text, Text)] }
     deriving (Show, Eq)
 newtype Title = Title { unTitle :: Html }
 
 newtype Head url = Head (HtmlUrl url)
     deriving Monoid
-instance Semigroup (Head a)
+instance Semigroup (Head url) where
+  (<>) = mappend
 newtype Body url = Body (HtmlUrl url)
     deriving Monoid
-instance Semigroup (Body a)
+instance Semigroup (Body url) where
+  (<>) = mappend
 
 type CssBuilderUrl a = (a -> [(Text, Text)] -> Text) -> TBuilder.Builder
 
@@ -374,25 +379,28 @@ data GWData a = GWData
     }
 instance Monoid (GWData a) where
     mempty = GWData mempty mempty mempty mempty mempty mempty mempty
-    mappend (GWData a1 a2 a3 a4 a5 a6 a7)
-            (GWData b1 b2 b3 b4 b5 b6 b7) = GWData
-        (a1 `mappend` b1)
-        (a2 `mappend` b2)
-        (a3 `mappend` b3)
-        (a4 `mappend` b4)
+#if !(MIN_VERSION_base(4,11,0))
+    mappend = (<>)
+#endif
+instance Semigroup (GWData a) where
+    GWData a1 a2 a3 a4 a5 a6 a7 <>
+      GWData b1 b2 b3 b4 b5 b6 b7 = GWData
+        (mappend a1 b1)
+        (mappend a2 b2)
+        (mappend a3 b3)
+        (mappend a4 b4)
         (unionWith mappend a5 b5)
-        (a6 `mappend` b6)
-        (a7 `mappend` b7)
-instance Semigroup (GWData a)
+        (mappend a6 b6)
+        (mappend a7 b7)
 
 data HandlerContents =
-      HCContent H.Status !TypedContent
-    | HCError ErrorResponse
-    | HCSendFile ContentType FilePath (Maybe FilePart)
-    | HCRedirect H.Status Text
-    | HCCreated Text
-    | HCWai W.Response
-    | HCWaiApp W.Application
+      HCContent !H.Status !TypedContent
+    | HCError !ErrorResponse
+    | HCSendFile !ContentType !FilePath !(Maybe FilePart)
+    | HCRedirect !H.Status !Text
+    | HCCreated !Text
+    | HCWai !W.Response
+    | HCWaiApp !W.Application
     deriving Typeable
 
 instance Show HandlerContents where
@@ -405,158 +413,81 @@ instance Show HandlerContents where
     show (HCWaiApp _) = "HCWaiApp"
 instance Exception HandlerContents
 
--- Instances for WidgetT
-instance Monad m => Functor (WidgetT site m) where
-    fmap = liftM
-instance Monad m => Applicative (WidgetT site m) where
-    pure = return
+-- Instances for WidgetFor
+instance Applicative (WidgetFor site) where
+    pure = WidgetFor . const . pure
     (<*>) = ap
-instance Monad m => Monad (WidgetT site m) where
-    return a = WidgetT $ const $ return (a, mempty)
-    WidgetT x >>= f = WidgetT $ \r -> do
-        (a, wa) <- x r
-        (b, wb) <- unWidgetT (f a) r
-        return (b, wa `mappend` wb)
-instance MonadIO m => MonadIO (WidgetT site m) where
-    liftIO = lift . liftIO
-instance MonadBase b m => MonadBase b (WidgetT site m) where
-    liftBase = WidgetT . const . liftBase . fmap (, mempty)
-instance MonadBaseControl b m => MonadBaseControl b (WidgetT site m) where
-#if MIN_VERSION_monad_control(1,0,0)
-    type StM (WidgetT site m) a = StM m (a, GWData (Route site))
-    liftBaseWith f = WidgetT $ \reader' ->
-        liftBaseWith $ \runInBase ->
-            fmap (\x -> (x, mempty))
-            (f $ runInBase . flip unWidgetT reader')
-    restoreM = WidgetT . const . restoreM
-#else
-    data StM (WidgetT site m) a = StW (StM m (a, GWData (Route site)))
-    liftBaseWith f = WidgetT $ \reader' ->
-        liftBaseWith $ \runInBase ->
-            fmap (\x -> (x, mempty))
-            (f $ fmap StW . runInBase . flip unWidgetT reader')
-    restoreM (StW base) = WidgetT $ const $ restoreM base
-#endif
-instance Monad m => MonadReader site (WidgetT site m) where
-    ask = WidgetT $ \hd -> return (rheSite $ handlerEnv hd, mempty)
-    local f (WidgetT g) = WidgetT $ \hd -> g hd
-        { handlerEnv = (handlerEnv hd)
-            { rheSite = f $ rheSite $ handlerEnv hd
-            }
-        }
+instance Monad (WidgetFor site) where
+    return = pure
+    WidgetFor x >>= f = WidgetFor $ \wd -> do
+        a <- x wd
+        unWidgetFor (f a) wd
+instance MonadIO (WidgetFor site) where
+    liftIO = WidgetFor . const
+-- | @since 1.4.38
+instance MonadUnliftIO (WidgetFor site) where
+  {-# INLINE askUnliftIO #-}
+  askUnliftIO = WidgetFor $ \wd ->
+                return (UnliftIO (flip unWidgetFor wd))
+instance MonadReader (WidgetData site) (WidgetFor site) where
+    ask = WidgetFor return
+    local f (WidgetFor g) = WidgetFor $ g . f
 
-instance MonadTrans (WidgetT site) where
-    lift = WidgetT . const . liftM (, mempty)
-instance MonadThrow m => MonadThrow (WidgetT site m) where
-    throwM = lift . throwM
+instance MonadThrow (WidgetFor site) where
+    throwM = liftIO . throwM
 
-instance MonadCatch m => MonadCatch (HandlerT site m) where
-  catch (HandlerT m) c = HandlerT $ \r -> m r `catch` \e -> unHandlerT (c e) r
-instance MonadMask m => MonadMask (HandlerT site m) where
-  mask a = HandlerT $ \e -> mask $ \u -> unHandlerT (a $ q u) e
-    where q u (HandlerT b) = HandlerT (u . b)
-  uninterruptibleMask a =
-    HandlerT $ \e -> uninterruptibleMask $ \u -> unHandlerT (a $ q u) e
-      where q u (HandlerT b) = HandlerT (u . b)
-instance MonadCatch m => MonadCatch (WidgetT site m) where
-  catch (WidgetT m) c = WidgetT $ \r -> m r `catch` \e -> unWidgetT (c e) r
-instance MonadMask m => MonadMask (WidgetT site m) where
-  mask a = WidgetT $ \e -> mask $ \u -> unWidgetT (a $ q u) e
-    where q u (WidgetT b) = WidgetT (u . b)
-  uninterruptibleMask a =
-    WidgetT $ \e -> uninterruptibleMask $ \u -> unWidgetT (a $ q u) e
-      where q u (WidgetT b) = WidgetT (u . b)
+instance MonadResource (WidgetFor site) where
+    liftResourceT f = WidgetFor $ runInternalState f . handlerResource . wdHandler
 
--- CPP to avoid a redundant constraints warning
-#if MIN_VERSION_base(4,9,0)
-instance (MonadIO m, MonadBase IO m, MonadThrow m) => MonadResource (WidgetT site m) where
-#else
-instance (Applicative m, MonadIO m, MonadBase IO m, MonadThrow m) => MonadResource (WidgetT site m) where
-#endif
-    liftResourceT f = WidgetT $ \hd -> liftIO $ (, mempty) <$> runInternalState f (handlerResource hd)
+instance MonadLogger (WidgetFor site) where
+    monadLoggerLog a b c d = WidgetFor $ \wd ->
+        rheLog (handlerEnv $ wdHandler wd) a b c (toLogStr d)
 
-instance MonadIO m => MonadLogger (WidgetT site m) where
-    monadLoggerLog a b c d = WidgetT $ \hd ->
-        liftIO $ (, mempty) <$> rheLog (handlerEnv hd) a b c (toLogStr d)
-
-#if MIN_VERSION_monad_logger(0, 3, 10)
-instance MonadIO m => MonadLoggerIO (WidgetT site m) where
-    askLoggerIO = WidgetT $ \hd -> return (rheLog (handlerEnv hd), mempty)
-#endif
-
-instance MonadActive m => MonadActive (WidgetT site m) where
-    monadActive = lift monadActive
-instance MonadActive m => MonadActive (HandlerT site m) where
-    monadActive = lift monadActive
-
-instance MonadTrans (HandlerT site) where
-    lift = HandlerT . const
+instance MonadLoggerIO (WidgetFor site) where
+    askLoggerIO = WidgetFor $ return . rheLog . handlerEnv . wdHandler
 
 -- Instances for HandlerT
-instance Monad m => Functor (HandlerT site m) where
-    fmap = liftM
-instance Monad m => Applicative (HandlerT site m) where
-    pure = return
+instance Applicative (HandlerFor site) where
+    pure = HandlerFor . const . return
     (<*>) = ap
-instance Monad m => Monad (HandlerT site m) where
-    return = HandlerT . const . return
-    HandlerT x >>= f = HandlerT $ \r -> x r >>= \x' -> unHandlerT (f x') r
-instance MonadIO m => MonadIO (HandlerT site m) where
-    liftIO = lift . liftIO
-instance MonadBase b m => MonadBase b (HandlerT site m) where
-    liftBase = lift . liftBase
-instance Monad m => MonadReader site (HandlerT site m) where
-    ask = HandlerT $ return . rheSite . handlerEnv
-    local f (HandlerT g) = HandlerT $ \hd -> g hd
-        { handlerEnv = (handlerEnv hd)
-            { rheSite = f $ rheSite $ handlerEnv hd
-            }
-        }
--- | Note: although we provide a @MonadBaseControl@ instance, @lifted-base@'s
--- @fork@ function is incompatible with the underlying @ResourceT@ system.
--- Instead, if you must fork a separate thread, you should use
--- @resourceForkIO@.
---
--- Using fork usually leads to an exception that says
--- \"Control.Monad.Trans.Resource.register\': The mutable state is being accessed
--- after cleanup. Please contact the maintainers.\"
-instance MonadBaseControl b m => MonadBaseControl b (HandlerT site m) where
-#if MIN_VERSION_monad_control(1,0,0)
-    type StM (HandlerT site m) a = StM m a
-    liftBaseWith f = HandlerT $ \reader' ->
-        liftBaseWith $ \runInBase ->
-            f $ runInBase . (\(HandlerT r) -> r reader')
-    restoreM = HandlerT . const . restoreM
-#else
-    data StM (HandlerT site m) a = StH (StM m a)
-    liftBaseWith f = HandlerT $ \reader' ->
-        liftBaseWith $ \runInBase ->
-            f $ fmap StH . runInBase . (\(HandlerT r) -> r reader')
-    restoreM (StH base) = HandlerT $ const $ restoreM base
-#endif
+instance Monad (HandlerFor site) where
+    return = pure
+    HandlerFor x >>= f = HandlerFor $ \r -> x r >>= \x' -> unHandlerFor (f x') r
+instance MonadIO (HandlerFor site) where
+    liftIO = HandlerFor . const
+instance MonadReader (HandlerData site site) (HandlerFor site) where
+    ask = HandlerFor return
+    local f (HandlerFor g) = HandlerFor $ g . f
 
-instance MonadThrow m => MonadThrow (HandlerT site m) where
-    throwM = lift . monadThrow
+-- | @since 1.4.38
+instance MonadUnliftIO (HandlerFor site) where
+  {-# INLINE askUnliftIO #-}
+  askUnliftIO = HandlerFor $ \r ->
+                return (UnliftIO (flip unHandlerFor r))
 
-instance (MonadIO m, MonadBase IO m, MonadThrow m) => MonadResource (HandlerT site m) where
-    liftResourceT f = HandlerT $ \hd -> liftIO $ runInternalState f (handlerResource hd)
+instance MonadThrow (HandlerFor site) where
+    throwM = liftIO . throwM
 
-instance MonadIO m => MonadLogger (HandlerT site m) where
-    monadLoggerLog a b c d = HandlerT $ \hd ->
-        liftIO $ rheLog (handlerEnv hd) a b c (toLogStr d)
+instance MonadResource (HandlerFor site) where
+    liftResourceT f = HandlerFor $ runInternalState f . handlerResource
 
-#if MIN_VERSION_monad_logger(0, 3, 10)
-instance MonadIO m => MonadLoggerIO (HandlerT site m) where
-    askLoggerIO = HandlerT $ \hd -> return (rheLog (handlerEnv hd))
-#endif
+instance MonadLogger (HandlerFor site) where
+    monadLoggerLog a b c d = HandlerFor $ \hd ->
+        rheLog (handlerEnv hd) a b c (toLogStr d)
+
+instance MonadLoggerIO (HandlerFor site) where
+    askLoggerIO = HandlerFor $ \hd -> return (rheLog (handlerEnv hd))
 
 instance Monoid (UniqueList x) where
     mempty = UniqueList id
-    UniqueList x `mappend` UniqueList y = UniqueList $ x . y
-instance Semigroup (UniqueList x)
+#if !(MIN_VERSION_base(4,11,0))
+    mappend = (<>)
+#endif
+instance Semigroup (UniqueList x) where
+    UniqueList x <> UniqueList y = UniqueList $ x . y
 
 instance IsString Content where
-    fromString = flip ContentBuilder Nothing . Blaze.ByteString.Builder.Char.Utf8.fromString
+    fromString = flip ContentBuilder Nothing . BB.stringUtf8
 
 instance RenderRoute WaiSubsite where
     data Route WaiSubsite = WaiSubsiteRoute [Text] [(Text, Text)]
@@ -580,3 +511,42 @@ data Logger = Logger
 
 loggerPutStr :: Logger -> LogStr -> IO ()
 loggerPutStr (Logger ls _) = pushLogStr ls
+
+-- | A handler monad for subsite
+--
+-- @since 1.6.0
+newtype SubHandlerFor sub master a = SubHandlerFor
+    { unSubHandlerFor :: HandlerData sub master -> IO a
+    }
+    deriving Functor
+
+instance Applicative (SubHandlerFor child master) where
+    pure = SubHandlerFor . const . return
+    (<*>) = ap
+instance Monad (SubHandlerFor child master) where
+    return = pure
+    SubHandlerFor x >>= f = SubHandlerFor $ \r -> x r >>= \x' -> unSubHandlerFor (f x') r
+instance MonadIO (SubHandlerFor child master) where
+    liftIO = SubHandlerFor . const
+instance MonadReader (HandlerData child master) (SubHandlerFor child master) where
+    ask = SubHandlerFor return
+    local f (SubHandlerFor g) = SubHandlerFor $ g . f
+
+-- | @since 1.4.38
+instance MonadUnliftIO (SubHandlerFor child master) where
+  {-# INLINE askUnliftIO #-}
+  askUnliftIO = SubHandlerFor $ \r ->
+                return (UnliftIO (flip unSubHandlerFor r))
+
+instance MonadThrow (SubHandlerFor child master) where
+    throwM = liftIO . throwM
+
+instance MonadResource (SubHandlerFor child master) where
+    liftResourceT f = SubHandlerFor $ runInternalState f . handlerResource
+
+instance MonadLogger (SubHandlerFor child master) where
+    monadLoggerLog a b c d = SubHandlerFor $ \sd ->
+        rheLog (handlerEnv sd) a b c (toLogStr d)
+
+instance MonadLoggerIO (SubHandlerFor child master) where
+    askLoggerIO = SubHandlerFor $ return . rheLog . handlerEnv
